"""
The FontFeatures class is a way of representing the transformations -
substitutions and positionings - going on inside a font at a semantically
high level. It aims to be independent of and unconstrained by the OpenType
representation, with the assumption that these high-level objects can be
either "compiled down" into AFDKO feature code or directly written to the
GSUB/GPOS tables of a font.

FontFeatures aims to marshal data between OTF binary representations,
AFDKO feature code, FontDame, and can power other representations such as the
FEZ language (see the 'fez' library).

A FontFeatures representation of a font will make use of two other top-level
concepts: Features and Routines. Routines are collections of rules; they play
a similar function to the AFDKO concept of a lookup, but unlike lookups,
Routines do not need to be comprised of rules of the same type. You can think
of them as functions that are called on a glyph string.

Here is an example of constructing a simple feature file using fontFeatures::

    ff = FontFeatures()

    liga_ffi = Substitution( [ ["f"], ["f"], ["i"] ], replacement=[["f_f_i"]] )
    liga_ffl = Substitution( [ ["f"], ["f"], ["l"] ], replacement=[["f_f_l"]] )
    liga_fi = Substitution( [ ["f"], ["i"] ], replacement=[["fi"]] )
    liga_ff = Substitution( [ ["f"], ["f"] ], replacement=[["f_f"]] )
    liga_routine = Routine(rules=[liga_ffi, liga_ffl, liga_fi, liga_ff])

    ff.addFeature("liga", [liga_routine])

    # Export Adobe syntax
    print(ff.asFea())

    font = TTFont("Test.ttf")
    ff.buildBinaryFeatures(font)
    font.save("Test-liga.ttf")
"""

from collections import OrderedDict, namedtuple
from fontTools.feaLib.ast import ValueRecord as feaLibValueRecord
from fontTools.feaLib.variableScalar import VariableScalar
from itertools import chain
from copy import copy


class FontFeatures:
    """An object representing the layout rules in a font.

    The initializer has no parameters."""

    def __init__(self):
        self.namedClasses = {}  #: A mapping of named classes to a list of glyph names which make up the class.
        self.routines = []  #: All of the layout routines used in this font.
        self.features = OrderedDict()  #: An ordered dictionary mapping feature tags to a list of routine references.
        self.anchors = {}  #: A dictionary mapping glyph names to a dictionary of anchor names / positions.
        self.symbols = {}
        self.glyphclasses = {}  #: A dictionary mapping glyph names to their categories.
        self.scratch = {}  #: Space for items to communicate context to each other.
        self.doneUsageMarking = False

    def __add__(self, other):
        """Combine two FontFeatures objects together."""
        combined = FontFeatures()
        for k in ["namedClasses", "routines", "features", "anchors", "symbols", "glyphclasses"]:
            if isinstance(getattr(combined, k), dict):
                getattr(combined, k).update(getattr(self, k))
                getattr(combined, k).update(getattr(other, k))
            else:
                setattr(combined, k, getattr(self, k) + getattr(other, k))
        return combined

    def gensym(self, category):
        """Generate a new unique symbol (used for labeling unlabeled data).

        Args:
            category (str): The category for this symbol

        Returns: a string representing a unique label."""
        if not category in self.symbols:
            self.symbols[category] = 0
        self.symbols[category] += 1
        return f"{category}{self.symbols[category]}"

    def routineNamed(self, name):
        """Finds a routine with the given name.

        Args:
            name (str): The name to find

        Returns: a :py:class:`Routine` object if the named routine was found
          in the features object. Raises a ``ValueError`` if not."""
        for r in self.routines:
            if r.name == name:
                return r
        raise ValueError("Can't find routine '%s'" % name)

    def referenceRoutine(self, r, do_usecount=True):
        """Store a routine and return a reference to it.

        Args:
            r: A :py:class:`Routine` object.
        """
        assert isinstance(r, Routine)

        if r not in self.routines:
            self.routines.append(r)
        r.parent = self
        if do_usecount:
            r.usecount = r.usecount + 1
        return RoutineReference(routine=r)

    def getNamedClassFor(self, glyphs, name):
        """Find and optionally stores a named class of glyphs

        Args:
            glyphs: A sequence of glyph names.
            name: A name for this glyph class if it does not exist.

        Returns:
            The name of a glyph class. If the exact same set of glyphs
            was already stored as a glyph class, then the name of that
            class will be returned. If not, then the class will be stored
            and the name provided as the ``name`` argument will be returned.
        """
        for k,v in self.namedClasses.items():
            if sorted(glyphs) == sorted(v):
                return k
        self.namedClasses[name] = glyphs
        return name

    def addFeature(self, name, rs):
        """Add Routines to a named feature.

        Args:
            name: The feature name.
            rs: A sequence of :py:class:`Routine` or :py:class:`RoutineReference` objects.
        """
        if not name in self.features:
            self.features[name] = []
        for r in rs:
            r.parent = self
            if isinstance(r, Routine):
                r = self.referenceRoutine(r)
            self.features[name].append(r)


    def allRules(self, ruletype=None):
        """Return all rules in the font, optionally filtered by type

        Args:
            ruletype: A class (``Positioning``, ``Substitution`` etc)
                to filter the results.

        Returns:
            Routines stored in the preamble and within features.
        """

        rules = []
        for r in self.routines:
            rules.extend(r.rules)

        if ruletype:
            rules = filter(lambda x: isinstance(x, ruletype), rules)
        return rules

    def markRoutineUseInChains(self):
        """Annotate routines which are used in chaining rules.

        Generally used when converting the fontFeatures object to another
        format; allows routines to know where they are being used by annotating
        them with the ``.usedin`` property for optimization purposes.
        """
        if self.doneUsageMarking:
            return
        for r in self.routines:
            r.usedin = set()
        for chain in self.allRules(Chaining):
            for routinelist in chain.lookups:
                if not routinelist:
                    continue
                for routine in routinelist:
                    # Using a set here so it is safe to call more than once
                    if isinstance(routine, RoutineReference):
                        routine.resolve(self)
                        routine.routine.usedin.add(chain)
                    else:
                        routine.usedin.add(chain)
        self.doneUsageMarking = True

    def hoist_languages(self):
        """Sort routines into scripts and languages and resolve wildcards."""
        scripts = OrderedDict()
        count = 0

        def add_language(p):
            nonlocal scripts
            nonlocal count
            s, l = p
            if not s in scripts:
                scripts[s] = []
            if l == "*":
                return
            if not l in scripts[s]:
                count = count + 1
                scripts[s].append(l)

        for k in self.routines:
            if k.languages:
                for l in k.languages:
                    add_language(l)
            else:
                for r in k.rules:
                    for l in (r.languages or []):
                        add_language(l)

        # if count > 0 and not "DFLT" in scripts:
        #     scripts["DFLT"] = []
        #     scripts.move_to_end("DFLT", last=False)
        # if count > 0 and not "dflt" in scripts["DFLT"]:
        #     scripts["DFLT"].insert(0, "dflt")

        self.scripts_and_languages = scripts

    def hasScriptSupport(self, script):
        """Check if the features object has support for a particular script.

        Args:
            script (str): A four-character OpenType script code.

        Returns: boolean
        """
        self.hoist_languages()
        return script in self.scripts_and_languages

    def resolveAllRoutines(self):
        """Resolve reference use in chains.

        Checks that all routines referenced in chain rules can actually
        be found within the object, and adds pointers to match named routine
        references with the relevant :py:class:`Routine` object.
        """
        for routine in self.routines:
            for r in routine.rules:
                if not isinstance(r, Chaining):
                    continue
                for lookuplistIx in range(len(r.lookups)):
                    r.lookups[lookuplistIx] = self.ensureLookupsAreReferences(
                        r.lookups[lookuplistIx]
                    )

    def ensureLookupsAreReferences(self, lookuplist):
        """Ensures that all references are lookups.

        Naughty people might put :py:class:`Routine` objects directly into
        :py:class:`Chain` lookups. This tidies them up."""
        rv = []
        if not lookuplist:
            return None
        for r in lookuplist:
            if isinstance(r, RoutineReference):
                r.resolve(self)
            else:
                r = self.referenceRoutine(r)
            assert r.routine in self.routines
            rv.append(r)
        return rv

    def setGlyphClassesFromFont(self, font):
        """Loads glyph classes from the font."""
        for g in font.exportedGlyphs():
            if hasattr(font, "glyphs"):
                self.glyphclasses[g] = font.glyphs[g].category
            else:
                self.glyphclasses[g] = font[g].category

    def partitionRoutine(self, routine, factor):
        """Splits a routine based on a predicate.

        This method applies a function to each rule in the routine and creates
        distinct routines, each containing rules with the same return value
        from the function. This is useful, for example, when exporting to
        OpenType, to ensure that all rules in a routine must have the same type,
        same flags, etc.

        Args:
            routine: A :py:class:`Routine` object.
            factor: A function applied to each of the :py:class:`Rule` objects.

        Returns: A list of :py:class:`Routine` objects. Additionally, modifies
        the ``.routines`` list of the FontFeatures object."""
        if not routine.rules:
            return
        self.doneUsageMarking = False
        self.markRoutineUseInChains()

        split_routines = {}
        split_rules = {}
        # The first rule stays in the original
        split_routines[factor(routine.rules[0])] = (routine, [routine.rules[0]])
        for rule in routine.rules[1:]:
            thisfactor = factor(rule)
            if thisfactor not in split_routines:
                rulelist = []
                newroutine = copy(routine)
                newroutine.rules = rulelist
                split_routines[thisfactor] = (newroutine, rulelist)
            split_routines[thisfactor][1].append(rule)

        # Alter the first rule's routine's rulelist (we couldn't alter it on
        # the fly because we were iterating over it).
        routine_with_first_rule, rulelist = split_routines[factor(routine.rules[0])]
        routine_with_first_rule.rules = rulelist

        allroutines = [x[0] for x in split_routines.values()]
        return self.replaceRoutineWithSplitList(routine, allroutines)

    def replaceRoutineWithSplitList(self, routine, allroutines):
        self.markRoutineUseInChains()
        index = self.routines.index(routine)
        usedin = routine.usedin
        self.routines[index : index + 1] = allroutines
        for user in usedin:
            for lookuplist in user.lookups:
                if lookuplist is None:
                    continue
                i = 0
                while i < len(lookuplist):
                    lookup = lookuplist[i]
                    if lookup == routine or (
                        isinstance(lookup, RoutineReference)
                        and lookup.routine == routine
                    ):
                        lookuplist[i : i + 1] = [
                            RoutineReference(routine=r) for r in allroutines
                        ]
                    i = i + 1
        # Same trick for features
        for lookuplist in self.features.values():
            if lookuplist is None:
                continue
            i = 0
            while i < len(lookuplist):
                lookup = lookuplist[i]
                if lookup == routine or (
                    isinstance(lookup, RoutineReference) and lookup.routine == routine
                ):
                    lookuplist[i : i + 1] = [
                        RoutineReference(routine=r) for r in allroutines
                    ]
                i = i + 1
        return allroutines

    from .feaLib.FontFeatures import asFea, asFeaAST
    from .xmlLib.FontFeatures import toXML, fromXML
    from .ttLib.FontFeatures import buildBinaryFeatures


class Routine:
    """Represent a Routine (similar to OT Lookup).

    A routine is a set of rules, sometimes but not always with an explicit name.
    It can apply to a set of language/script pairs.
    """

    def __init__(
        self,
        name="",
        rules=None,
        address=None,
        inlined=False,
        languages=None,
        parent=None,
        flags=0,
        markFilteringSet=None,
        markAttachmentSet=None,
    ):
        self.name = name
        self.usecount = 0
        self.usedin = set()
        if rules:
            self.rules = rules
        else:
            self.rules = []
        self.address = address
        self.comments = []
        self.inlined = inlined
        self.languages = languages or []
        self.parent = parent
        self.flags = flags
        self.markFilteringSet = markFilteringSet
        self.markAttachmentSet = markAttachmentSet

    def addRule(self, rule):
        """Adds a rule to a Routine.

        Args:
            rule: A ``Substitution``, ``Positioning``, etc. object.
        """
        assert isinstance(rule, Rule)
        self.rules.append(rule)

    def addComment(self, comment):
        """Adds a comment to a Routine.

        Comments are emitted when the Routine is converted to text formats
        such as AFDKO.

        Args:
            comment: A string comment.
        """
        self.comments.append(comment)

    @property
    def involved_glyphs(self):
        """Returns the names of all of the glyphs involved in this Routine."""
        return set().union(*(r.involved_glyphs for r in self.rules))

    @property
    def stage(self):
        """Returns which shaping stage this routine is used in.

        Returns: ``sub`` for substitution stage, ``pos`` for positioning stage."""
        for r in self.rules:
            if isinstance(r, Substitution):
                return "sub"
            if isinstance(r, Positioning):
                return "pos"
            if isinstance(r, Attachment):
                return "pos"
            if isinstance(r, Chaining):
                return r.stage

    @property
    def dependencies(self):
        """Returns a list of :py:class:`Routine` objects called as lookups in
        this Routine."""
        deps = []
        for r in self.rules:
            deps.extend(r.dependencies)
        return deps

    from .feaLib.Routine import asFea, asFeaAST, feaPreamble
    from .shaperLib.Routine import apply_to_buffer
    from .xmlLib.Routine import toXML, fromXML
    from .ttLib.Routine import toOTLookup


class ExtensionRoutine(Routine):
    """OpenType-specific concept: A routine which contains other routines."""
    def __init__(self, **kwargs):
        if "routines" in kwargs:
            self.routines = kwargs["routines"]
            del kwargs["routines"]
        super().__init__(**kwargs)

    def apply_to_buffer(self, buf, stage=None, feature=None):
        """Applies shaping rules from this routine to a buffer.

        Args:
            buf: A :py:class:`fontFeatures.shaperLib.Buffer` object.
            stage (str): Shaping stage - ``sub`` or ``pos``.
            feature (str): The feature being processed. (For debugging.)

        Modifies the ``buf`` object.
        """
        for r in self.routines:
            r.apply_to_buffer(buf, stage, feature)

    def asFeaAST(self):
        """Returns this extension routine as ``fontTools.feaLib.ast`` objects."""
        import fontTools.feaLib.ast as feaast

        f = feaast.Block()
        for r in self.routines:
            f.statements.append(r.asFeaAST())
        return f

    @property
    def stage(self):
        """Returns which shaping stage this routine is used in.

        Returns: ``sub`` for substitution stage, ``pos`` for positioning stage."""
        return self.routines[0].stage

    @property
    def rules(self):
        """All rules under this extension.

        Returns: A flattened list of :py:class:`Rule` objects."""
        return list(chain(*[routine.rules for routine in self.routines]))

    @rules.setter
    def rules(self, foo):
        """Does nothing. Don't set rules here."""
        pass


class RoutineReference:
    """A reference to a Routine object, used in a lookup.

    Routines can be referenced either by name (for example, when loaded from a
    textual representation), in which case they will be resolved at a later time,
    or by providing a pointer to the :py:class:`Routine` object."""
    def __init__(self, name=None, routine=None):
        self.routine = routine
        if self.routine:
            self.name = routine.name
        if name:
            self.name = name

    def resolve(self, ff):
        """Resolves the reference in the context of a :py:class:`FontFeatures`
        object.

        Raises a ``ValueError`` if a named routine cannot be found.
        """
        if not self.routine:
            self.routine = ff.routineNamed(self.name)
            if not self.routine:
                raise ValueError("Could not resolve routine")
        if not self.name:
            self.name = self.routine.name

    @property
    def stage(self):
        """Returns which shaping stage this routine is used in.

        Returns: ``sub`` for substitution stage, ``pos`` for positioning stage."""
        if not self.routine:
            raise ValueError("Routine reference not resolved")
        return self.routine.stage

    from .xmlLib.RoutineReference import fromXML, toXML
    from .feaLib.RoutineReference import asFeaAST, feaPreamble

    def asFea(self):
        """Returns this Rule as a string of AFDKO feature text."""
        return self.asFeaAST().asFea()

class Rule:
    """A base class for all rules."""
    def asFea(self):
        """Returns this Rule as a string of AFDKO feature text."""
        return self.asFeaAST().asFea()

    def feaPreamble(self, ff):
        """Computes any text that needs to go in the feature file header."""
        return []

    from .shaperLib.Rule import would_apply_at_position, pre_post_context_matches
    from .xmlLib.Rule import fromXML, toXML, _makeglyphslots, _slotArray

    @property
    def has_context(self):
        """Does this rule have any pre- or post-context defined?"""
        return len(self.precontext) or len(self.postcontext)

    @property
    def dependencies(self):
        """Returns a list of :py:class:`Routine` objects called as lookups in
        this Routine."""
        return []


class Substitution(Rule):
    """Represents a Substitution rule.

    A substitution represents any kind of exchange of one set of glyphs for
    another: single substitutions, multiple substitutions and ligatures are all
    substitutions. Optionally, substitutions may be followed by precontext and
    postcontext.

    Args:
        input_: A list of lists. The outer list represents the positions in
            the glyph stream to substitute, with the inner list representing
            the glyph names at each position.
        replacement: A list of glyph names.
        precontext: A list of list of glyphs which must appear before the input
            sequence.
        postcontext: A list of list of glyphs which must appear before the input
            sequence.
        lookups: A list of list of lookups to be applied to the glyph sequence.
            The outer list represents the positions in the input sequence, with
            the inner list containing Routines to apply.
        reverse: Boolean representing if the substitutions should take place from
            the end of the string.
        force_alt: Force this substitution to be interpreted as an alternate
            substitution.

    Examples::

        lig = Substitution(
            [ ["f"], ["i"] ],
            ["f_i"]
        ) # sub f i by f_i;

        contextual = Substitution(
            [ ["dotbelow"] ],
            [ ["dotbelow.post"] ],
            precontext = [["ra-myanmar", "ra-myanmar.bt1", "ra-myanmar.bt2"]]
        ) # sub [ra-myanmar ra-myanmar.bt1 ra-myanmar.bt2] dotbelow-myanmar'
          # by dotbelow-myanmar.post;
    """

    def __init__(
        self,
        input_,
        replacement,
        precontext=None,
        postcontext=None,
        address=None,
        languages=None,
        lookups=None,
        reverse=False,
        flags=0,
        force_alt=False
    ):
        self.precontext = precontext or []
        self.postcontext = postcontext or []
        self.input = input_
        self.replacement = replacement
        self.address = address
        self.lookups = lookups or []
        self.languages = languages
        self.flags = flags
        self.reverse = reverse
        self.stage = "sub"
        self.force_alt = force_alt

    @property
    def involved_glyphs(self):
        """Returns a set of all glyphs involved in this rule."""
        i = set(chain.from_iterable(self.input))
        o = set(chain.from_iterable(self.replacement))
        b = set(chain.from_iterable(self.precontext))
        a = set(chain.from_iterable(self.postcontext))
        return i | o | b | a

    from .feaLib.Substitution import asFeaAST
    from .shaperLib.Substitution import shaper_inputs, _do_apply
    from .xmlLib.Substitution import _toXML, fromXML
    from .ttLib.Substitution import lookup_type


class Chaining(Rule):
    """Represents a Chain rule.

    A Chain rule represents the operation of calling another Routine when
    a particular input context is met.

    Args:
        input_: A list of lists. The outer list represents the positions in
            the glyph stream to substitute, with the inner list representing
            the glyph names at each position.
        precontext: A list of list of glyphs which must appear before the input
            sequence.
        postcontext: A list of list of glyphs which must appear before the input
            sequence.
        lookups: A list of list of lookups to be applied to the glyph sequence.
            The outer list represents the positions in the input sequence, with
            the inner list containing Routines to apply.

    Example::

        sub_Qu = Routine(rules=[
            Substitute([["Q"]], [["Q.beforeu"]])
        ])

        chain = Chain(
            [["Q"]],
            postcontext = [ ["u", "v", "u.sc", "v.sc"] ],
            lookups = [ [sub_Qu] ]
        ) # sub Q' lookup sub_Qu [u v u.sc v.sc];
    """

    def __init__(
        self,
        input_,
        precontext=None,
        postcontext=None,
        address=None,
        languages=None,
        lookups=None,
        flags=0,
    ):
        self.precontext = precontext or []
        self.postcontext = postcontext or []
        self.input = input_
        self.address = address
        self.lookups = lookups or []
        self.languages = languages
        self.flags = flags

    @property
    def stage(self):
        """Returns which shaping stage this routine is used in.

        Returns: ``sub`` for substitution stage, ``pos`` for positioning stage."""
        for l in self.lookups:
            if not l:
                continue
            for aLookup in l:
                if not aLookup:
                    continue
                return aLookup.stage

    @property
    def involved_glyphs(self):
        """Returns a set of all glyphs involved in this rule."""
        i = set(chain.from_iterable(self.input))
        b = set(chain.from_iterable(self.precontext))
        a = set(chain.from_iterable(self.postcontext))
        return i | b | a

    @property
    def dependencies(self):
        """Returns a list of :py:class:`Routine` objects called as lookups in
        this Routine."""
        deps = []
        for l in self.lookups:
            for aLookup in (l or []):
                if isinstance(aLookup, RoutineReference):
                    deps.append(aLookup.routine)
                else:
                    deps.append(aLookup)
        return deps

    from .feaLib.Chaining import asFeaAST, feaPreamble
    from .shaperLib.Chaining import shaper_inputs, _do_apply
    from .xmlLib.Chaining import _toXML, fromXML
    from .ttLib.Chaining import lookup_type


class ValueRecord(feaLibValueRecord):
    """A value record for representing positional changes in advance and placement.

    See :py:class:`fontTools.feaLib.ValueRecord`, from which this inherits.
    """
    from .ttLib.ValueRecord import toOTValueRecord

    @property
    def is_variable(self):
        """Returns true if any of the elements of the value record are a
        :py:class:`fontTools.feaLib.VariableScalar`."""
        return any(
            isinstance(x, VariableScalar)
            for x in [
                self.xPlacement,
                self.yPlacement,
                self.xAdvance,
                self.yAdvance,
            ]
        )


class Positioning(Rule):
    """Represents a Positioning rule.

    Args:
        input_: A list of lists. The outer list represents the positions in
            the glyph stream to position, with the inner list representing
            the glyph names at each glyph stream position.
        valuerecords: A list of ``ValueRecord`` objects to be applied at each
            glyph stream position.
        precontext: A list of list of glyphs which must appear before the input
            sequence.
        postcontext: A list of list of glyphs which must appear before the input
            sequence.

    Example::

        open_up_behs = Positioning(
            [
                ["BEi1", "BEi2"],
                ["sda", "sdb", "dda", "ddb"]
            ],
            [
                ValueRecord(xAdvance=200),
                ValueRecord(xPlacement=50),
            ]
            postcontext = [ medis_finas ]
        )
        # pos [BEi1 BEi2]' <0 0 200 0> [sda sdb dda ddb]' <0 50 0 0> @medis_finas;
    """

    def __init__(
        self,
        glyphs,
        valuerecords,
        precontext=None,
        postcontext=None,
        address=None,
        languages=None,
        flags=0,
    ):
        self.precontext = precontext or []
        self.postcontext = postcontext or []
        assert len(glyphs) == len(valuerecords)
        self.glyphs = glyphs
        self.valuerecords = valuerecords
        self.address = address
        self.languages = languages
        self.flags = flags
        self.stage = "pos"

    @property
    def involved_glyphs(self):
        """Returns a set of all glyphs involved in this rule."""
        i = set(chain.from_iterable(self.glyphs))
        b = set(chain.from_iterable(self.precontext))
        a = set(chain.from_iterable(self.postcontext))
        return i | b | a

    from .feaLib.Positioning import asFeaAST
    from .shaperLib.Positioning import shaper_inputs, _do_apply
    from .xmlLib.Positioning import _toXML, fromXML
    from .ttLib.Positioning import lookup_type


class Attachment(Rule):
    """Represents an Attachment rule.

    Args:
        base_name: Name of the base class.
        mark_name: Name of the mark class.
        bases: Dictionary. They keys are names of glyphs to act as bases to
            the attachment (this may be categorized as mark glyphs if the
            attachment is a mark-to-mark operation); the associated values are
            a two-element tuple with the coordinates of the anchor.
        marks: Dictionary. They keys are names of glyphs to act as marks;
            the associated values are a two-element tuple with the coordinates
            of the anchor.
        force_markmark: boolean. If true, force this to be interpreted as a
            mark-to-mark operation

    Whether this is a mark-to-base or mark-to-mark operation will be determined
    by the glyph category of the glyphs involved in the `bases` dictionary and
    the value of the `force_markmark` argument.

    Examples::

        ff.anchors = {
            "a": { "top": (250, 603) },
            "acutecomb": { "_top": (56, 0) }
        }

        top_bases = {}
        top_marks = {}
        for glyphname, anchors in ff.anchors.items():
            for anchorname, position in anchors.items():
                if anchorname == "top":
                    top_bases[glyphname] = position
                if anchorname == "_top":
                    top_marks[glyphname] = position

        # top_bases = { "a": (260,603) }
        # top_marks = { "acutecomb": (56,0) }

        tops = Attachment("top", "_top", top_bases, top_marks)
    """

    def __init__(
        self,
        base_name,
        mark_name,
        bases=None,
        marks=None,
        flags=0,
        address=None,
        font=None,
        languages=None,
        force_markmark=False
    ):
        self.base_name = base_name
        self.mark_name = mark_name
        self.bases = bases or {}
        self.marks = marks or {}
        self.flags = flags
        self.address = address
        self.font = font
        self.stage = "pos"
        self.languages = languages or []
        self.force_markmark = force_markmark

    @property
    def is_cursive(self):
        """Returns true if this is a cursive attachment rule."""
        return self.base_name == "cursive_entry" or self.base_name == "entry"  # XXX

    from .feaLib.Attachment import asFeaAST, feaPreamble
    from .shaperLib.Attachment import shaper_inputs, _do_apply, would_apply_at_position
    from .xmlLib.Attachment import _toXML, fromXML
    from .ttLib.Attachment import lookup_type

    @property
    def involved_glyphs(self):
        """Returns a set of all glyphs involved in this rule."""
        b = set(self.bases.keys())
        m = set(self.marks.keys())
        return b | m



from fontFeatures import FontFeatures, Routine
from lxml import etree
import re


expected = """
    <fontfeatures>
        <namedclasses>
            <class name="One">
                <glyph>A</glyph>
                <glyph>B</glyph>
                <glyph>C</glyph>
            </class>
            <class name="Two">
                <glyph>d</glyph>
                <glyph>e</glyph>
                <glyph>f</glyph>
            </class>
        </namedclasses>
        <routines>
            <routine name="One"/>
            <routine name="Two"/>
        </routines>
        <features>
            <feature name="onex">
                <routinereference name="One"/>
            </feature>
            <feature name="twox">
                <routinereference name="Two"/>
            </feature>
        </features>
        <anchors>
            <glyph name="a">
                <anchor name="top" x="300" y="200"/>
            </glyph>
        </anchors>
        <glyphclasses>
            <glyph name="a" class="base"/>
            <glyph name="d" class="mark"/>
        </glyphclasses>
    </fontfeatures>"""
expected = re.sub(r"(?m)^\s+", "", expected).replace("\n", "")


def test_add():
    r1 = Routine(name="One")
    r2 = Routine(name="Two")
    f1 = FontFeatures()
    f2 = FontFeatures()
    f1.namedClasses["One"] = ["A", "B", "C"]
    f1.glyphclasses["a"] = "base"
    f1.addFeature("onex", [r1])
    f2.namedClasses["Two"] = ["d", "e", "f"]
    f2.addFeature("twox", [r2])
    f2.glyphclasses["d"] = "mark"
    f2.anchors["a"] = { "top": (300, 200) }

    combined = f1 + f2
    assert combined.namedClasses["One"] == ["A", "B", "C"]
    assert combined.namedClasses["Two"] == ["d", "e", "f"]
    assert "onex" in combined.features
    assert combined.features["onex"][0].routine == r1
    assert "twox" in combined.features
    assert combined.features["twox"][0].routine == r2
    assert combined.glyphclasses["a"] == "base"
    assert combined.glyphclasses["d"] == "mark"

    el = combined.toXML()
    assert etree.tostring(el).decode() == expected

def test_fromXML():
    ff = FontFeatures.fromXML(etree.fromstring(expected))
    assert ff.namedClasses["One"] == ["A", "B", "C"]
    assert ff.namedClasses["Two"] == ["d", "e", "f"]
    assert "onex" in ff.features
    assert ff.features["onex"][0].routine == ff.routines[0]
    assert "twox" in ff.features
    assert ff.features["twox"][0].routine == ff.routines[1]
    assert ff.glyphclasses["a"] == "base"
    assert ff.glyphclasses["d"] == "mark"
    assert ff.anchors["a"] == { "top": (300, 200) }


def test_routine_named():
    r1 = Routine(name="One")
    f1 = FontFeatures()
    f1.addFeature("onex", [r1])
    assert f1.routineNamed("One") == r1



"""Routines for converting Routine objects to and from XML."""
from lxml import etree
from fontFeatures.xmlLib.utils import put_languages, put_address


def toXML(self):
    """Serializes a Routine to a lxml Element object."""
    root = etree.Element("routine")
    if self.flags:
        root.attrib["flags"] = str(self.flags)
    put_address(self, root)
    put_languages(self, root)
    if self.name:
        root.attrib["name"] = self.name
    for r in self.rules:
        root.append(r.toXML())

    return root


@classmethod
def fromXML(klass, el):
    """Creates a Routine from a lxml Element object."""
    from fontFeatures import Rule

    rule = klass(
        address=(el.get("address") or "").split("|"), name=el.get("name"), flags=(int(el.get("flags") or 0))
    )
    for r in el:
        rule.addRule(Rule.fromXML(r))
    return rule



# Code for converting a Attachment object into feaLib statements
import fontTools.feaLib.ast as feaast
from glyphtools import categorize_glyph
import warnings
from fontTools.feaLib.variableScalar import VariableScalar


def fix_scalar(scalar):
    if isinstance(scalar, int):
        return scalar
    if isinstance(scalar, float):
        return int(scalar)
    if hasattr(scalar, "axes"):
        # assume it's a variable scalar of some kind
        return scalar
    raise ValueError("Illegal anchor position %s" % scalar)

def _glyphref(g):
    if len(g) == 1:
        return feaast.GlyphName(g[0])
    return feaast.GlyphClass([feaast.GlyphName(x) for x in g])


def sortByAnchor(self):
    anchors = {}
    for name, pos in self.marks.items():
        if not pos in anchors:
            anchors[pos] = []
        anchors[pos].append(name)
    self.markslist = [(v, k) for k, v in anchors.items()]

    anchors = {}
    for name, pos in self.bases.items():
        if not pos in anchors:
            anchors[pos] = []
        anchors[pos].append(name)
    self.baseslist = [(v, k) for k, v in anchors.items()]


def feaPreamble(self, ff):
    if self.is_cursive:
        return []
    sortByAnchor(self)
    if not "mark_classes_done" in ff.scratch:
        ff.scratch["mark_classes_done"] = {}
    b = feaast.Block()
    fullname = self.base_name + "_" + self.mark_name
    for mark in self.markslist:
        if not (fullname, tuple(mark[0])) in ff.scratch["mark_classes_done"]:
            b.statements.append(
                feaast.MarkClassDefinition(
                    feaast.MarkClass(fullname),
                    feaast.Anchor(fix_scalar(mark[1][0]), fix_scalar(mark[1][1])),
                    _glyphref(mark[0]),
                )
            )
            ff.scratch["mark_classes_done"][(fullname, tuple(mark[0]))] = True
    return [b]


def asFeaAST(self):
    b = feaast.Block()
    if self.is_cursive:
        allglyphs = set(self.bases.keys()) | set(self.marks.keys())
        for g in allglyphs:
            b.statements.append(
                feaast.CursivePosStatement(
                    _glyphref([g]),
                    g in self.bases and feaast.Anchor(fix_scalar(self.bases[g][0]),fix_scalar(self.bases[g][1])),
                    g in self.marks and feaast.Anchor(fix_scalar(self.marks[g][0]),fix_scalar(self.marks[g][1])),
                )
            )
    else:
        if not hasattr(self, "baseslist"):
            sortByAnchor(self)  # e.g. when testing
        for base in self.baseslist:
            statementtype = feaast.MarkBasePosStatement
            fullname = self.base_name + "_" + self.mark_name
            if self.font:
                if categorize_glyph(self.font, base[0][0])[0] == "mark":
                    statementtype = feaast.MarkMarkPosStatement
            if self.force_markmark:
                statementtype = feaast.MarkMarkPosStatement
            b.statements.append(
                statementtype(
                    _glyphref(base[0]),
                    [[feaast.Anchor(fix_scalar(base[1][0]), fix_scalar(base[1][1])), feaast.MarkClass(fullname)]],
                )
            )

    return b



"""
GTableUnparser: Base class for reading binary GSUB/GPOS tables
==============================================================
"""
from collections import OrderedDict
from fontTools.misc.xmlWriter import XMLWriter
import fontFeatures
from io import BytesIO
import warnings


def glyph(x):
    """Helper routine to document that a glyph name goes in a slot."""
    assert isinstance(x, str)
    return [x]


class GTableUnparser:
    """Base class for reading binary GSUB/GPOS tables."""

    def __init__(self, table, ff, languageSystems, font=None, config={}):
        """Create a new unparser

        Args:
            table: An object returned by ``font["GSUB"]`` or ``font["GPOS"]``.
            ff: A fontFeatures object in which to return the output.
            font: A ``TTFont`` object (optional).
            config: A dictionary mapping generated names to friendly names.

        The config dictionary is used to make the output more human readable.
        For example, after dumping a binary TTF and reading the output, you might
        discover that what we are calling ``SingleSubstitution1`` would be better
        named ``fiLigature``. Passing ``{"SingleSubstitution1": "fiLigature"}``
        to the call will cause the routine to be renamed appropriately."""

        self.table = table.table
        self.font = font
        self.fontFeatures = ff
        self.config = config
        self.index = 0
        self.lookups = []  # We keep a separate list because this is per-table
        self.sharedClasses = {}
        self.languageSystems = languageSystems
        self.sharedLookups = OrderedDict()

    def _unparse_lookups(self, slr, inputs, in_lookups=None):
        lookups = []
        note = None
        if in_lookups:
            lookups = in_lookups
        indices = [sl.SequenceIndex for sl in slr]
        if indices != list(sorted(indices)):
            # https://github.com/adobe-type-tools/afdko/issues/1167
            note = (
                f"Out-of-order lookup application in lookup {self.currentLookup}\n"
                f" A chaining rule calls lookups in the order {indices}\n"
                " This cannot be directly expressed in Adobe FEA syntax;\n"
                "  asFea() output will be incorrect"
            )
            warnings.warn(note)
        for sl in slr:
            if len(lookups) <= sl.SequenceIndex:
                lookups.extend([None] * (1 + sl.SequenceIndex - len(lookups)))
            if not lookups[sl.SequenceIndex]:
                lookups[sl.SequenceIndex] = []
            if sl.LookupListIndex >= len(self.lookups):
                raise ValueError("Lookups unparsed out of order: unparse %i first!" % sl.LookupListIndex)
            rr = fontFeatures.RoutineReference(routine=self.lookups[sl.LookupListIndex])
            lookups[sl.SequenceIndex].append(rr)
        if len(lookups) < len(inputs):
            lookups.extend([None] * (len(inputs) - len(lookups)))
        assert len(lookups) == len(inputs)
        return lookups, note

    def _invertClassDef(self, a, font):
        classes = {}
        for glyph, klass in a.items():
            if klass not in classes:
                classes[klass] = []
            classes[klass].append(glyph)
        glyphset = set(font.getGlyphOrder())
        classes[0] = glyphset - set(a.keys())
        return classes

    def getname(self, n):
        """Renames its input using the configuration dictionary"""
        return self.config.get(n, n)

    def gensym(self):
        """Generates a new unique name"""
        self.index = self.index + 1
        return str(self.index)

    def unparse(self, doLookups=True):
        """Unparse the table to the fontFeatures object."""
        if not self.table.ScriptList:
            return
        self.unparseLookups()
        self.collectFeatures()
        self.fontFeatures.resolveAllRoutines()

    def _prepareFeatureLangSys(self, langTag, langSys, table, features, scriptTag):
        # This is a part of prepareFeatures
        for featureIdx in langSys.FeatureIndex:
            featureRecord = self.table.FeatureList.FeatureRecord[featureIdx]
            featureTag = featureRecord.FeatureTag
            scripts = features.get(featureTag, None)
            if scripts is None:
                scripts = OrderedDict()
                features[featureTag] = scripts

            languages = scripts.get(scriptTag, None)
            if languages is None:
                languages = OrderedDict()
                scripts[scriptTag] = languages

            lookups = languages.get(langTag, None)
            if lookups is None:
                lookups = []
                languages[langTag] = lookups

            for lookupIdx in featureRecord.Feature.LookupListIndex:
                self.lookups[lookupIdx].languages.append((scriptTag, langTag))
                # Add reference if there isn't one
                if not featureTag in self.fontFeatures.features:
                    self.fontFeatures.features[featureTag] = []
                if not any(
                    r.routine == self.lookups[lookupIdx]
                    for r in self.fontFeatures.features[featureTag]
                ):
                    self.fontFeatures.addFeature(
                        featureTag,
                        [
                            self.lookups[lookupIdx]
                        ],
                    )
                lookups.append(lookupIdx)

    def collectFeatures(self):
        """Traverse the script/language table looking for features."""
        features = OrderedDict()
        for scriptRecord in self.table.ScriptList.ScriptRecord:
            scriptTag = scriptRecord.ScriptTag
            if scriptRecord.Script.DefaultLangSys is not None:
                self._prepareFeatureLangSys(
                    "dflt",
                    scriptRecord.Script.DefaultLangSys,
                    self.table,
                    features,
                    scriptTag,
                )
            for langSysRecord in scriptRecord.Script.LangSysRecord:
                self._prepareFeatureLangSys(
                    langSysRecord.LangSysTag,
                    langSysRecord.LangSys,
                    self.table,
                    features,
                    scriptTag,
                )
        self.features = features

    def addFeatures(self, doLookups=True):
        """Add lookup references from the features list into the fontFeatures object."""
        for name, feature in self.features.items():
            f = []
            for scriptname, langs in feature.items():
                for lang, lookups in langs.items():
                    if doLookups:
                        for lookupIdx in lookups:
                            f.append(
                                RoutineReference(
                                    routine=self.fontFeatures.routines[lookupIdx]
                                )
                            )
                            self.fontFeatures.routines[lookupIdx].languages.append(
                                (scriptname, lang)
                            )
            self.fontFeatures.addFeature(name, f)

    def unparseLookups(self):
        """Unparses the lookups to fontFeatures routines."""
        if not self.table.LookupList:
            return
        # Create a dummy list first, to allow resolving chained lookups
        for _ in self.table.LookupList.Lookup:
            r = fontFeatures.Routine()
            self.lookups.append(r)
            self.fontFeatures.routines.append(r)

        for lookupIdx, lookup in enumerate(self.table.LookupList.Lookup):
            routine, deps = self.unparseLookup(lookup, lookupIdx)
            debug = self.getDebugInfo(self._table, lookupIdx)
            if debug:
                routine.address = (self._table, lookupIdx, *debug)
                if debug[1]:
                    routine.name = debug[1]
            self._copyRoutineToRoutine(routine, self.lookups[lookupIdx])

    def _copyRoutineToRoutine(self, src, dst):
        dst.name = src.name
        dst.rules = src.rules
        dst.flags = src.flags
        dst.address = src.address
        dst.comments = src.comments
        dst.inlined = src.inlined
        dst.languages = src.languages
        dst.parent = src.parent
        dst.flags = src.flags
        dst.markFilteringSet = src.markFilteringSet
        dst.markAttachmentSet = src.markAttachmentSet

    def unparseLookup(self, lookup, lookupIdx):
        """Dispatches to the appropriate lookup unparser."""
        self.currentLookup = lookupIdx
        unparser = getattr(self, "unparse" + self.lookupTypes[lookup.LookupType])
        return unparser(lookup)

    def unparseExtension(self, lookup):
        """Handles extension lookups by recursing into them."""
        routines = []
        dependencies = []
        for xt in lookup.SubTable:
            xt.SubTable = [xt.ExtSubTable]
            xt.LookupType = xt.ExtSubTable.LookupType
            xt.LookupFlag = lookup.LookupFlag
            if hasattr(lookup, "MarkFilteringSet"):
                xt.MarkFilteringSet = lookup.MarkFilteringSet
            routine, deps = self.unparseLookup(xt, self.currentLookup)
            routines.append(routine)
            dependencies.extend(deps)
        extension = fontFeatures.ExtensionRoutine(
            routines=routines,
            name=self.getname("Extension" + self.gensym()),
        )
        self._fix_flags(extension, lookup)
        return extension, dependencies

    def getDebugInfo(self, table, ix):
        """If information is present in the font's Debg table, use that to
        retrieve the original address and routine names."""
        if not self.font or "Debg" not in self.font:
            return None
        debug_data = self.font["Debg"].data
        if "com.github.fonttools.feaLib" not in debug_data:
            return None
        debug_data = debug_data["com.github.fonttools.feaLib"][table][str(ix)]
        return debug_data[0], debug_data[1]

    def _asXML(self, sub):
        writer = XMLWriter(BytesIO())
        sub.toXML(writer, self.font)
        out = writer.file.getvalue().decode("utf-8")
        return out

    def unparsable(self, b, e, sub):
        """Warn about an unparsable lookup."""
        import warnings

        warnings.warn(
            "# XXX Unparsable rule: " + str(e) + " in " + str(self.currentLookup)
        )
        b.addComment("# ----")
        out = self._asXML(sub).splitlines()
        for ln in out:
            b.addComment("# " + ln)
        b.addComment("# ----\n")

    def _fix_flags(self, routine, lookup):
        routine.flags = lookup.LookupFlag
        mat = lookup.LookupFlag & 0xFF00
        if mat:
            mat = mat >> 8
            classDefs = self.font["GDEF"].table.MarkAttachClassDef.classDefs
            glyphs = [g for g in classDefs.keys() if classDefs[g] == mat]
            routine.markAttachmentSet = glyphs
        if lookup.LookupFlag & 0x10 and hasattr(lookup, "MarkFilteringSet"):
            routine.markFilteringSet = (
                self.font["GDEF"]
                .table.MarkGlyphSetsDef.Coverage[lookup.MarkFilteringSet]
                .glyphs
            )

        if routine.flags & 0x10:
            assert routine.markFilteringSet is not None


    def unparseContextual(self, lookup):
        """Handles a generic contextual lookup, in various formats."""
        b = fontFeatures.Routine(
            name=self.getname("Contextual" + self._table + self.gensym())
        )
        self._fix_flags(b, lookup)
        for sub in lookup.SubTable:
            if sub.Format == 1:
                self._unparse_contextual_format1(sub, b, lookup)
            elif sub.Format == 2:
                self._unparse_contextual_format2(sub, b, lookup)
            elif sub.Format == 3:
                self._unparse_contextual_format3(sub, b, lookup)
            else:
                raise ValueError
        return b, []

    def _unparse_contextual_format1(self, sub, b, lookup):
        lookups = []
        rulesetattr, ruleattr, lookupattr = [
            self._attrs[x] for x in ["format1_ruleset", "format1_rule", "lookup"]
        ]

        for subrulesets, input_ in zip(getattr(sub, rulesetattr), sub.Coverage.glyphs):
            for subrule in getattr(subrulesets, ruleattr):
                lookups = []
                allinput = [glyph(x) for x in ([input_] + subrule.Input)]
                lookups, note = self._unparse_lookups(getattr(subrule, lookupattr), allinput)
                rule = fontFeatures.Chaining(
                        allinput,
                        lookups=lookups,
                        address=self.currentLookup,
                        flags=lookup.LookupFlag,
                    )
                if note:
                    rule.note = note
                b.addRule(rule)
        return

    def _unparse_contextual_format2(self, sub, b, lookup):
        return self._unparse_contextual_chain_format2(sub, b, lookup, chain=False)

    def _unparse_contextual_format3(self, sub, b, lookup):
        return self._unparse_contextual_chain_format3(sub, b, lookup, chain=False)

    def unparseChainedContextual(self, lookup):
        """Handles a generic chained contextual lookup, in various formats."""
        b = fontFeatures.Routine(
            name=self.getname("ChainedContextual" + self._table + self.gensym())
        )
        self._fix_flags(b, lookup)
        for sub in lookup.SubTable:
            if sub.Format == 1:
                self._unparse_contextual_chain_format1(sub, b, lookup)
            elif sub.Format == 2:
                self._unparse_contextual_chain_format2(sub, b, lookup)
            elif sub.Format == 3:
                self._unparse_contextual_chain_format3(sub, b, lookup)
            else:
                raise ValueError
        return b, []

    def _unparse_contextual_chain_format1(self, sub, b, lookup):
        rulesetattr, ruleattr, lookupattr = [
            self._attrs[x]
            for x in ["chain_format1_ruleset", "chain_format1_rule", "lookup"]
        ]

        for subrulesets, input_ in zip(getattr(sub, rulesetattr), sub.Coverage.glyphs):
            for subrule in getattr(subrulesets, ruleattr):
                lookups = []
                inputs = [glyph(x) for x in ([input_] + subrule.Input)]
                prefix = [glyph(x) for x in reversed(subrule.Backtrack)]
                suffix = [glyph(x) for x in subrule.LookAhead]
                lookups, note = self._unparse_lookups(getattr(subrule, lookupattr), inputs)
                rule = fontFeatures.Chaining(
                        inputs,
                        prefix,
                        suffix,
                        lookups=lookups,
                        address=self.currentLookup,
                        flags=lookup.LookupFlag,
                    )
                if note:
                    rule.note = note
                b.addRule(rule)

    def _unparse_contextual_chain_format2(self, sub, b, lookup, chain=True):
        if chain:
            rulesetattr, ruleattr, lookupattr = [
                self._attrs[x]
                for x in ["chain_format2_classset", "chain_format2_rule", "lookup"]
            ]
        else:
            rulesetattr, ruleattr, lookupattr = [
                self._attrs[x] for x in ["format2_classset", "format2_rule", "lookup"]
            ]

        if chain:
            backtrack = {}
            if sub.BacktrackClassDef:
                backtrack = self._invertClassDef(
                    sub.BacktrackClassDef.classDefs, self.font
                )
            lookahead = {}
            if sub.LookAheadClassDef:
                lookahead = self._invertClassDef(
                    sub.LookAheadClassDef.classDefs, self.font
                )
            inputs = {}
            inputs = self._invertClassDef(sub.InputClassDef.classDefs, self.font)
        else:
            inputs = self._invertClassDef(sub.ClassDef.classDefs, self.font)

        rulesets = getattr(sub, rulesetattr)
        for classId, ruleset in enumerate(rulesets):
            if not ruleset:
                continue
            rules = getattr(ruleset, ruleattr)
            inputclass = inputs.get(classId, [])
            # The coverage filters the input class...
            inputclass = [g for g in inputclass if g in sub.Coverage.glyphs]
            for r in rules:
                if chain:
                    prefix = list(reversed([backtrack[x] for x in r.Backtrack]))
                    suffix = [lookahead[x] for x in r.LookAhead]
                    input_ = [inputclass] + [inputs[x] for x in r.Input]
                else:
                    prefix, suffix = [], []
                    input_ = [inputclass] + [inputs[x] for x in r.Class]
                lookups, note = self._unparse_lookups(getattr(r, lookupattr), input_)
                rule = fontFeatures.Chaining(
                        input_,
                        prefix,
                        suffix,
                        lookups=lookups,
                        address=self.currentLookup,
                        flags=lookup.LookupFlag,
                    )
                if note:
                    rule.note = note
                b.addRule(rule)

    def _unparse_contextual_chain_format3(self, sub, b, lookup, chain=True):
        lookupattr = self._attrs["lookup"]
        prefix, suffix, inputs = [], [], []

        if chain:
            for coverage in reversed(sub.BacktrackCoverage):
                prefix.append(coverage.glyphs)
            for coverage in sub.LookAheadCoverage:
                suffix.append(coverage.glyphs)
            for coverage in sub.InputCoverage:
                inputs.append(coverage.glyphs)
        else:
            for coverage in sub.Coverage:
                inputs.append(coverage.glyphs)

        lookups, note = self._unparse_lookups(getattr(sub, lookupattr), inputs)
        rule = fontFeatures.Chaining(
                inputs,
                prefix,
                suffix,
                lookups=lookups,
                address=self.currentLookup,
                flags=lookup.LookupFlag,
            )
        if note:
            rule.note = note
        b.addRule(rule)



from fontFeatures import Substitution, FontFeatures
from fontTools.ttLib import TTFont
from fontFeatures.ttLib.GSUBUnparser import GSUBUnparser
from fontFeatures.ttLib import unparseLanguageSystems
import pprint
import unittest


class TestUnparse(unittest.TestCase):
    font = TTFont("fonts/Amiri-Regular.ttf")
    lookups = font["GSUB"].table.LookupList.Lookup
    ff = FontFeatures()
    unparser = GSUBUnparser(font["GSUB"], ff, [])

    def test_single(self):
        g, _ = self.unparser.unparseLookup(self.lookups[1], 1)  # part of locl
        self.assertEqual(g.rules[0].asFea(), "sub period by period.ara;")
        self.assertEqual(g.rules[1].asFea(), "sub guillemotleft by guillemotleft.ara;")

    def test_ligature(self):
        g, _ = self.unparser.unparseLookup(self.lookups[0], 0)  # part of ccmp
        self.assertEqual(g.rules[0].asFea(), "sub uni0627 uni065F by uni0673;")

    def test_multiple(self):
        g, _ = self.unparser.unparseLookup(self.lookups[10], 10)
        self.assertEqual(g.rules[0].asFea(), "sub uni08B6 by uni0628 smallmeem.above;")

    def test_ignore(self):
        g, _ = self.unparser.unparseLookup(self.lookups[47], 47)
        self.unparser.lookups = [None] * 47 + [g]
        assert self.unparser.lookups[47]
        g, _ = self.unparser.unparseLookup(self.lookups[48], 48)
        self.assertEqual(
            g.rules[0].asFea(),
            "ignore sub [uni0622 uni0627 uni0648 uni0671 uni0627.fina uni0671.fina] uni0644.init' uni0644.medi' [uni0647.fina uni06C1.fina];",
        )

    # def test_chaining(self):
    #     self.unparser.unparseLookups()
    #     g, _ = self.unparser.unparseLookup(
    #         self.lookups[33], 33
    #     )  # part of calt in quran.fea
    #     self.assertEqual(
    #         g.rules[0].asFea(),
    #         "sub uni0644' lookup SingleSubstitution32 uni0621' lookup SingleSubstitution31 uni0627' lookup SingleSubstitution32;",
    #     )



from dataclasses import dataclass
from fontFeatures import ValueRecord
from glyphtools import get_glyph_metrics, categorize_glyph
from youseedee import ucd_data
import sys
import warnings


def _add_value_records(vr1, vr2):
    if vr1.xPlacement or vr2.xPlacement:
        vr1.xPlacement = (vr1.xPlacement or 0) + (vr2.xPlacement or 0)
    if vr1.yPlacement or vr2.yPlacement:
        vr1.yPlacement = (vr1.yPlacement or 0) + (vr2.yPlacement or 0)
    if vr1.xAdvance or vr2.xAdvance:
        vr1.xAdvance = (vr1.xAdvance or 0) + (vr2.xAdvance or 0)
    if vr1.yAdvance or vr2.yAdvance:
        vr1.yAdvance = (vr1.yAdvance or 0) + (vr2.yAdvance or 0)


@dataclass
class BufferItem:
    # codepoint: int
    # glyph: str
    # position: ValueRecord
    # category: str

    def __repr__(self):
        s = ""
        if self.glyph:
            s = self.glyph
            if self.category[0] == "base":
                s = s + "_"
            elif self.category[0] == "mark":
                s = s + "^"
            elif self.category[0] == "ligature":
                s = s + "(ﬁ)"
            else:
                s = s + "?"
        else:
            s = "U+%04x" % self.codepoint
        return "BufferItem(%s)" % s

    @classmethod
    def new_unicode(klass, codepoint):
        self = klass()
        self.codepoint = codepoint
        self.glyph = None
        self.feature_masks = {}
        return self

    @classmethod
    def new_glyph(klass, glyph, font):
        self = klass()
        self.codepoint = None
        self.glyph = glyph
        self.feature_masks = {}
        self.prep_glyph(font)
        return self

    def map_to_glyph(self, font):
        if not self.glyph:
            self.glyph = font.unicode_map.get(self.codepoint)
        if not self.glyph:
            # Notdef
            self.glyph = list(font.glyphs.keys())[0]
        self.prep_glyph(font)

    def prep_glyph(self, font):
        if "pytest" in sys.modules:
            if self.glyph in font.exportedGlyphs():
                self.gid = font.exportedGlyphs().index(self.glyph)
            else:
                self.gid = -1 # ?
        self.substituted = False
        self.ligated = False
        self.multiplied = False
        self.recategorize(font)
        try:
            self.position = ValueRecord(xAdvance=0)
            self.position.xAdvance=font.default_master.get_glyph_layer(self.glyph).width
        except Exception as e:
            if "pytest" in sys.modules:
                # We tolerate broken fonts in pytest
                pass
            else:
                raise ValueError("Could not get xAdvance for glyph %s (%i)" % (self.glyph, self.codepoint))

    def recategorize(self, font):
        try:
            self.category = categorize_glyph(font, self.glyph)
            if not self.category[0]:
                self.category = ("unknown", None)
        except Exception as e:
            warnings.warn("Error getting category: %s" % str(e))
            self._fallback_categorize()

    def _fallback_categorize(self):
        if not self.codepoint:
            # Now what?
            self.category = ("unknown", None)
            return
        genCat = ucd_data(self.codepoint).get("General_Category", "L")
        if genCat[0] == "M":
            self.category = ("mark", None)
        elif genCat == "Ll":
            self.category = ("ligature", None)
        elif genCat[0] == "L":
            self.category = ("base", None)
        else:
            self.category = ("unknown", None)

    def add_position(self, vr2):
        _add_value_records(self.position, vr2)


class Buffer:
    """A buffer holding either characters to be shaped or shaped glyphs."""
    itemclass = BufferItem

    def __init__(self, font, glyphs=[], unicodes=[], direction=None, script=None, language=None):
        self.font = font
        self.direction = direction
        self.script = script
        self.language = language
        self.fallback_mark_positioning = False
        self.fallback_glyph_classes = False
        self.items = []
        self.mask = []
        self.flags = 0
        self.current_feature_mask = None
        if glyphs:
            self.store_glyphs(glyphs)
            self.clear_mask()
        elif unicodes:
            self.store_unicode(unicodes)
            self.guess_segment_properties()

    def store_glyphs(self, glyphs):
        """Initialize the buffer with list of glyphs.

        Args:
            glyphs (list): A list of glyph names.
        """
        self.items = [self.itemclass.new_glyph(g, self.font) for g in glyphs]

    def store_unicode(self, unistring):
        """Initialize the buffer with list of Unicode codepoints.

        Args:
            glyphs (list or str): A list of characters.
        """
        self.items = [self.itemclass.new_unicode(ord(char)) for char in unistring ]

    def guess_segment_properties(self):
        """Try to automatically determine the script and direction properties."""
        for u in self.items:
            # Guess segment properties
            if not self.script:
                thisScript = ucd_data(u.codepoint)["Script"]
                if thisScript not in ["Common", "Unknown", "Inherited"]:
                    self.script = thisScript
        if not self.direction:
            from fontFeatures.shaperLib.Shaper import _script_direction
            self.direction = _script_direction(self.script)

    def map_to_glyphs(self):
        """Convert a buffer of codepoints to its initial glyph mapping."""
        glyphs = []
        for u in self.items:
            u.map_to_glyph(self.font)
        self.clear_mask()

    @property
    def is_all_glyphs(self):
        """Returns true if the contents are glyph items."""
        return all([x.glyph is not None for x in self.items])

    @property
    def is_all_unicodes(self):
        """Returns true if the contents are character items."""
        return all([x.codepoint is not None for x in self.items])

    def __getitem__(self, key):
        indexed = self.mask[key]
        if isinstance(indexed, range) or isinstance(indexed, slice):
            indexed = slice(indexed.start, indexed.stop, indexed.step)
        if isinstance(indexed, list):
            return [self.items[g] for g in indexed]
        return self.items[indexed]

    def __setitem__(self, key, value):
        indexed = self.mask[key]
        if len(indexed) == 1:  # Easy
            self.items[indexed[0] : indexed[0] + 1] = value
            return
        if len(value) == 1:  # Also easy
            self.items[indexed[0]] = value[0]
            for i in reversed(indexed[1:]):
                del self.items[i]
            return
        else:
            raise ValueError("Too hard :-(")

    def __len__(self):
        return len(self.mask)

    def update(self):
        """Categorises glyphs and recomputes masks.

        Called internally when the contents of the buffer changes."""
        for g in self.items:
            g.recategorize(self.font)
        self.recompute_mask()

    def clear_mask(self):
        """Clear the buffer mask."""
        self.flags = 0
        self.markFilteringSet = None
        self.markAttachmentSet = None
        self.current_feature_mask = None
        self.recompute_mask()

    def set_mask(self, flags, markFilteringSet=None, markAttachmentSet=None):
        """Apply a routine's flags and mark filtering set to the buffer."""
        self.flags = flags
        if self.flags & 0x10:
            assert(markFilteringSet)
        self.markFilteringSet = markFilteringSet
        if self.flags & 0xFF00:
            assert(markAttachmentSet)
        self.markAttachmentSet = markAttachmentSet
        self.recompute_mask()

    def recompute_mask(self):
        """Computes the mask property

        ``buffer.mask`` is a list of buffer item indices which is the current
        "view" of the buffer. For example, if ``buffer.flags == 0x8``, then the
        indices of all mark glyphs will be excluded from ``buffer.mask``.
        """
        mask = range(0, len(self.items))
        self.flags = self.flags or 0
        if self.flags & 0x2:  # IgnoreBases
            mask = list(filter(lambda ix: self.items[ix].category[0] != "base", mask))
        if self.flags & 0x4:  # IgnoreLigatures
            mask = list(
                filter(lambda ix: self.items[ix].category[0] != "ligature", mask)
            )
        if self.flags & 0x8:  # IgnoreMarks
            mask = list(filter(lambda ix: self.items[ix].category[0] != "mark", mask))
        if self.flags & 0x10:  # UseMarkFilteringSet
            mask = list(
                filter(
                    lambda ix: self.items[ix].category[0] != "mark"
                    or self.items[ix].glyph in self.markFilteringSet,
                    mask,
                )
            )
        if self.flags & 0xFF00:  # MarkAttachmentType
            mask = list(
                filter(
                    lambda ix: self.items[ix].category[0] != "mark"
                    or self.items[ix].glyph in self.markAttachmentSet,
                    mask,
                )
            )

        if self.current_feature_mask:
            feature = self.current_feature_mask
            mask = list(
                filter(
                    lambda ix: (feature not in self.items[ix].feature_masks)
                        or (not self.items[ix].feature_masks[feature]),
                    mask
                )
            )
        self.mask = mask

    def set_feature_mask(self, feature):
        """Applies the mask for a particular feature."""
        self.current_feature_mask = feature
        self.recompute_mask()

    def move_item(self, src, dest):
        """Moves an item from src to dest."""
        self.items[dest:dest] = [ self.items.pop(src) ]

    def merge_clusters(self, start, end):
        """Currently unimplemented."""
        pass  # XXX

    def serialize(self, additional = None, position=True, names=True, ned=False):
        """Serialize a buffer to a string.

    Returns:
        The contents of the given buffer in a string format similar to
        that used by ``hb-shape``.

    """
        outs = []
        if additional:
            if not isinstance(additional, list):
                additional = [additional]
        else:
            additional = []
        xcursor = 0
        for ix,info in enumerate(self.items):
            if hasattr(info, "glyph") and info.glyph:
                if names:
                    outs.append("%s" % info.glyph)
                else:
                    outs.append("%s" % info.gid)
            else:
                outs.append("U+%04x" % info.codepoint)
            if ned:
                position = info.position
                if xcursor + (position.xPlacement or 0):
                    outs[-1] = outs[-1] + "@%i,%i" % (xcursor + (position.xPlacement or 0), position.yPlacement or 0)
                xcursor = xcursor + position.xAdvance
            elif position and hasattr(info, "position"):
                position = info.position
                if hasattr(info, "syllable_index"):
                    cluster = info.syllable_index
                else:
                    cluster = ix
                outs[-1] = outs[-1] + "=%i" % (cluster)
                if position.xPlacement or position.yPlacement:
                    outs[-1] = outs[-1] + "@%i,%i" % (
                        position.xPlacement or 0,
                        position.yPlacement or 0,
                    )
                outs[-1] = outs[-1] + "+%i" % (position.xAdvance)
            relevant = list(filter(lambda a: hasattr(info,a), additional))
            if relevant:
                outs[-1] = outs[-1] + "(%s)" % ",".join(
                    [str(getattr(info, a)) for a in relevant]
                )
        return "|".join(outs)



#!/usr/bin/env python
import logging
import os
import sys
import re
from argparse import ArgumentParser
from collections import OrderedDict

import fontTools.feaLib.ast as feaast
import glyphsLib
from fontTools.ttLib import TTFont
from glyphsLib.classes import GSAnchor, GSClass, GSFeature, GSFeaturePrefix, Point
from lxml import etree

import fontFeatures
from fontFeatures.feaLib.FontFeatures import add_language_system_statements
from fontFeatures.ttLib import unparse
from fontFeatures.ttLib.GDEFUnparser import GDEFUnparser

LOGLEVEL = os.environ.get("LOGLEVEL", "WARNING").upper()
logging.basicConfig(level=LOGLEVEL)

import warnings


def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
    return "# [warning] %s\n" % (message)


warnings.formatwarning = warning_on_one_line

parser = ArgumentParser()
parser.add_argument("source", help="source file to process", metavar="SOURCE")
parser.add_argument("firstmaster", help="binary font file to process", metavar="TTF")
parser.add_argument("--no-kerning", help="Don't try to convert kerning", action="store_true")
parser.add_argument(
    "-o",
    "--output",
    dest="output",
    help="Output source file (defaults to overwriting original)",
)
parser.add_argument(
    "--config", default=None, help="config file to process", metavar="CONFIG"
)
parser.add_argument("othermasters", help="additional binary masters", metavar="TTF", nargs="*")

args = parser.parse_args()

config = {}
if args.config:
    import json

    with open(args.config) as f:
        config = json.load(f)

class Layout2Glyphs:
    def __init__(self, gsfont, ff, args, master=0):
        self.gsfont = gsfont
        self.ff = ff
        self.master = master
        self.args = args

    def process(self):
        self.process_gpos()

        self.fix_gdef_classes()
        self.add_named_classes()
        self.add_routines_to_feature_prefix()
        self.ensure_mark_mkmk()
        self.add_features_to_glyphs()
        self.add_language_systems()
        self.fix_duplicate_anchors()

    def process_gpos(self):
        for feature in ["mark", "mkmk", "abvm", "blwm", "dist"]:
            if feature in self.ff.features:
                self.add_anchors(feature)
        for feature in ["kern", "dist"]:
            if feature in ff.features and not args.no_kerning:
                self.add_kerning(feature)


    def add_anchor_to_glyph(self, glyphname, anchor_name, pos):
        anchor = GSAnchor(name=anchor_name, position=Point(*pos))
        glyph = self.gsfont.glyphs[glyphname]
        glyph.layers[self.master].anchors.append(anchor)
        return anchor

    def add_anchors(self, feature):
        mark = self.ff.features[feature]
        new_mark = []
        for rr in mark:
            routine = rr.routine
            delete_me = False
            if "MarkToLigature" in routine.name:
                self.add_mark_lig_anchors(routine.name, routine.comments[2:-1])
                delete_me = True
            else:
                delete_me = self.add_anchors_routine(routine)
            if not delete_me:
                new_mark.append(rr)
            else:
                del self.ff.routines[self.ff.routines.index(routine)]
        self.ff.features[feature] = new_mark

    def add_anchors_routine(self, routine):
        delete_me = False
        for rule in routine.rules:
            if isinstance(rule, fontFeatures.Attachment):
                delete_me = True
                if "Cursive" in routine.name:
                    anchor_name = "entry"
                else:
                    anchor_name = rule.base_name
                for base, pos in rule.bases.items():
                    self.add_anchor_to_glyph(base, anchor_name, pos)
                if "Cursive" in routine.name:
                    anchor_name = "exit"
                else:
                    anchor_name = "_"+rule.base_name
                for mark, pos in rule.marks.items():
                    self.add_anchor_to_glyph(mark, anchor_name, pos)
        return delete_me

    def add_kern(self, left, right, val):
        old_val = self.gsfont.kerningForPair(
            self.gsfont.masters[self.master].id,
            left,
            right,
        )
        if old_val == self.gsfont.EMPTY_KERNING_VALUE:
            old_val = 0
        else:
            print("Adding %i to existing kern %s/%s=%i" % (val, left, right, old_val))
        if not val:
            return
        self.gsfont.setKerningForPair(
            self.gsfont.masters[self.master].id,
            left,
            right,
            old_val + val
        )

    def add_kerning(self, feature):
        kern = self.ff.features[feature]
        new_kern = []
        for rr in kern:
            routine = rr.routine
            delete_me = False
            pairs = set()
            for rule in routine.rules:
                if isinstance(rule, fontFeatures.Positioning):
                    # Is it pair positioning
                    if len(rule.glyphs) != 2:
                        break
                    # Is it a simple kern?
                    if not (rule.valuerecords[0] and not rule.valuerecords[1]):
                        continue
                    if not rule.valuerecords[0].xAdvance:
                        continue
                    if rule.valuerecords[0].xPlacement or rule.valuerecords[0].yPlacement:
                        continue
                    for left in rule.glyphs[0]:
                        for right in rule.glyphs[1]:
                            if (left, right) in pairs:
                                continue
                            pairs.add((left,right))
                            self.add_kern(left, right, rule.valuerecords[0].xAdvance)
                            print(
                                "# Kerning %s %s = %i"
                                % (left, right, rule.valuerecords[0].xAdvance)
                            )
                    delete_me = True

            if not delete_me:
                new_kern.append(rr)
            elif routine in ff.routines:
                del ff.routines[ff.routines.index(routine)]
        ff.features[feature] = new_kern

    def add_mark_lig_anchors(self, name, xml):
        xml = "\n".join([re.sub(r"#\s+", "", c) for c in xml])
        tree = etree.fromstring(xml)
        marks = [x.get("value") for x in tree.xpath("//*/MarkCoverage/Glyph")]
        ligs = [x.get("value") for x in tree.xpath("//*/LigatureCoverage/Glyph")]
        array = tree.xpath("//*/LigatureArray")[0]
        for ligature, attach in zip(ligs, array.iter("LigatureAttach")):
            for ix, cr in enumerate(attach.iter("ComponentRecord")):
                for markclass, anchor in enumerate(cr.iter("LigatureAnchor")):
                    x = int(anchor.xpath(".//XCoordinate")[0].get("value"))
                    y = int(anchor.xpath(".//YCoordinate")[0].get("value"))
                    anchor_name = name+f"_class{markclass}_"+str(1 + ix)
                    self.add_anchor_to_glyph(ligature, anchor_name, (x,y))

        array = tree.xpath("//*/MarkArray")[0]
        for mark, attach in zip(marks, array.iter("MarkRecord")):
            markclass = int(attach.find("Class").get("value"))
            anchor_name = "_"+name+f"_class{markclass}"
            x = int(attach.xpath(".//XCoordinate")[0].get("value"))
            y = int(attach.xpath(".//YCoordinate")[0].get("value"))
            anchor = self.add_anchor_to_glyph(mark, anchor_name, (x,y))

    def fix_gdef_classes(self):
        for glyph, cat in self.ff.glyphclasses.items():
            if cat == "mark":
                self.gsfont.glyphs[glyph].category = "Mark"
                if self.gsfont.glyphs[glyph].layers[0].width == 0:
                    self.gsfont.glyphs[glyph].subCategory = "Nonspacing"
                else:
                    self.gsfont.glyphs[glyph].subCategory = "Spacing Combining"
            if cat == "ligature":
                self.gsfont.glyphs[glyph].category = "Letter"
                self.gsfont.glyphs[glyph].subCategory = "Ligature"

    def add_named_classes(self):
        ff.asFea()
        for cls, members in self.ff.namedClasses.items():
            self.gsfont.classes.append(GSClass(name=cls, code=" ".join(members)))

    def add_feature_prefix(self, routine):
        self.gsfont.featurePrefixes.append(
            GSFeaturePrefix(name=routine.name, code=routine.asFea())
        )

    def add_routines_to_feature_prefix(self):
        # Lookups go into prefix, in order of dependency
        queue = list(self.ff.routines)
        seen = {}
        while queue:
            for dep in queue[0].dependencies:
                if dep not in seen:
                    queue.insert(0, dep)
            routine = queue.pop(0)
            if routine not in seen:
                self.add_feature_prefix(routine)
            seen[routine] = True

    def ensure_mark_mkmk(self):
        # Work around https://github.com/googlefonts/ufo2ft/issues/506
        if "mark" not in self.ff.features and "mkmk" in self.ff.features:
            newfeatures = OrderedDict()
            for k, v in self.ff.features.items():
                if k == "mkmk":
                    newfeatures["mark"] = []
                newfeatures[k] = v
            self.ff.features = newfeatures

    def add_features_to_glyphs(self):
        for featuretag, routines in self.ff.features.items():
            code = "\n".join([
                f"# {routine.languages}\n"+
                routine.asFea() for routine in routines
            ])
            if featuretag in ["mark", "mkmk", "dist", "kern", "abvm", "blwm"]:
                code = "# Automatic Code Start\n" + code
            self.gsfont.features.append(GSFeature(name=featuretag, code=code))

    def add_language_systems(self):
        code = ""
        for script, langs in self.ff.scripts_and_languages.items():
            for lang in langs:
                code += f"languagesystem {script} {lang};\n"
        self.gsfont.featurePrefixes.append(GSFeaturePrefix(name="Languagesystems", code=code))

    def fix_duplicate_anchors(self):
        #  Get names of all anchors
        glyphs_with_anchor = {}
        for g in self.gsfont.glyphs:
            for l in g.layers:
                for a in l.anchors:
                    glyphs_with_anchor.setdefault(a.name, set()).add(g.name)

        our_anchors = [a for a in glyphs_with_anchor.keys() if "Anchor" in a]
        their_anchors = [a for a in glyphs_with_anchor.keys() if "Anchor" not in a]
        drop_anchors = set()
        if their_anchors:
            for theirs in their_anchors:
                for ours in our_anchors:
                    if glyphs_with_anchor[theirs] != glyphs_with_anchor[ours]:
                        continue
                    same = True
                    for g in glyphs_with_anchor[theirs]:
                        for l in self.gsfont.glyphs[g].layers:
                            if not l.anchors[theirs]:
                                print("Bad anchor %s in %s" % (theirs, l))
                                same = False
                            elif not l.anchors[ours]:
                                print("Bad anchor %s in %s" % (ours, l))
                                same = False
                            elif l.anchors[theirs].position != l.anchors[ours].position:
                                same = False
                    if not same:
                        continue
                    print("Our anchor '%s' was duplicate of original source '%s'" % (ours, theirs))
                    drop_anchors.add(ours)
                    break

        for to_drop in drop_anchors:
            for glyph in glyphs_with_anchor[to_drop]:
                for l in self.gsfont.glyphs[glyph].layers:
                    del l.anchors[to_drop]

gsfont = glyphsLib.load(open(args.source))

font = TTFont(args.firstmaster)

ff = unparse(font, do_gdef=True, doLookups=True, config=config)

Layout2Glyphs(gsfont, ff, args).process()

for ix, master in enumerate(args.othermasters):
    print("%s -> %i" % (master, ix+1))
    font = TTFont(master)
    ff = unparse(font, do_gdef=True, doLookups=True, config=config)
    Layout2Glyphs(gsfont, ff, args, master=ix+1).process_gpos()

if not args.output:
    args.output = args.source
gsfont.save(args.output)
print(f"Saved {args.output}")



"""Routines for converting Substitution rules to and from XML."""
from lxml import etree

def _toXML(self, root):
  self._makeglyphslots(root, "from", self.input)
  self._makeglyphslots(root, "to", self.replacement)
  return root

@classmethod
def fromXML(klass, el):
  """Creates a rule from a lxml Element object."""
  rule = klass(
      klass._slotArray(klass, el.find("from")),
      klass._slotArray(klass, el.find("to")),
      precontext = klass._slotArray(klass, el.find("precontext")),
      postcontext = klass._slotArray(klass, el.find("postcontext")),
      address = el.get("address"),
      languages = el.get("languages"),
      reverse = el.get("reverse"),
      flags = el.get("flags")
  )
  if el.find("lookups"):
    rule.lookups = []
    for slot in list(el):
      routines = [Routine.fromXML(x) for x in slot.findall("routine")]
      rule.lookups.append(routines)
  return rule



#!env python
import sys
from fontFeatures.feaLib import FeaUnparser
from fontFeatures.optimizer import Optimizer
from argparse import ArgumentParser

import logging
import os

LOGLEVEL = os.environ.get("LOGLEVEL", "WARNING").upper()
logging.basicConfig(level=LOGLEVEL)

import warnings


def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
    return "# [warning] %s\n" % (message)


warnings.formatwarning = warning_on_one_line

parser = ArgumentParser()
parser.add_argument("input", help="feature file to process", metavar="FILE")
parser.add_argument(
    "--optimize", dest="optimize", action="store_true", help="Run optimizer"
)

args = parser.parse_args()
ff = FeaUnparser(open(args.input, "r")).ff
if args.optimize:
    Optimizer().optimize(ff)
print(ff.asFea())



"""
voltLib: Load fontFeatures objects from Microsoft VOLT
======================================================

This module is experimental and incomplete.
"""

import logging
import re
from io import StringIO
from fontTools.ttLib import TTFont, TTLibError
from fontTools.voltLib import ast as VAst
from fontTools.voltLib.parser import Parser
from fontFeatures import ValueRecord, FontFeatures, Routine, Substitution, RoutineReference

log = logging.getLogger()

class Group:
    def __init__(self, group):
        self.name = group.name.lower()
        self.groups = [
            x.group.lower() for x in group.enum.enum if isinstance(x, VAst.GroupName)
        ]

    def __lt__(self, other):
        if self.name in other.groups:
            return True
        if other.name in self.groups:
            return False
        return self.name < other.name


class VoltParser:
    _NOT_LOOKUP_NAME_RE = re.compile(r"[^A-Za-z_0-9.]")
    _NOT_CLASS_NAME_RE = re.compile(r"[^A-Za-z_0-9.\-]")

    def __init__(self, file_or_path, font=None):
        self._file_or_path = file_or_path
        self._font = font

        self._glyph_map = {}
        self._glyph_order = None

        self._gdef = {}
        self._glyphclasses = {}
        self._features = {}
        self._lookups = {}

        self._marks = set()
        self._ligatures = {}

        self._markclasses = {}
        self._anchors = {}

        self._settings = {}

        self._lookup_names = {}
        self._class_names = {}

    def _lookupName(self, name):
        if name not in self._lookup_names:
            res = self._NOT_LOOKUP_NAME_RE.sub("_", name)
            while res in self._lookup_names.values():
                res += "_"
            self._lookup_names[name] = res
        return self._lookup_names[name]

    def _className(self, name):
        if name not in self._class_names:
            res = self._NOT_CLASS_NAME_RE.sub("_", name)
            while res in self._class_names.values():
                res += "_"
            self._class_names[name] = res
        return self._class_names[name]

    def _collectStatements(self, doc):
        # Collect and sort group definitions first, to make sure a group
        # definition that references other groups comes after them since VOLT
        # does not enforce such ordering, and feature file require it.
        groups = [s for s in doc.statements if isinstance(s, VAst.GroupDefinition)]
        for statement in sorted(groups, key=lambda x: Group(x)):
            self._groupDefinition(statement)

        for statement in doc.statements:
            if isinstance(statement, VAst.GlyphDefinition):
                self._glyphDefinition(statement)
            elif isinstance(statement, VAst.AnchorDefinition):
                self._anchorDefinition(statement)
            elif isinstance(statement, VAst.SettingDefinition):
                self._settingDefinition(statement)
            elif isinstance(statement, VAst.GroupDefinition):
                pass  # Handled above
            elif isinstance(statement, VAst.ScriptDefinition):
                self._scriptDefinition(statement)
            elif not isinstance(statement, VAst.LookupDefinition):
                raise NotImplementedError(statement)

        # Lookup definitions need to be handled last as they reference glyph
        # and mark classes that might be defined after them.
        for statement in doc.statements:
            if isinstance(statement, VAst.LookupDefinition):
                self._lookupDefinition(statement)

        # Now rearrange features
        for feature, thing in self._features.items():
            for script, thing in thing.items():
                for lang, lookups in thing.items():
                    for lookup in lookups:
                        ref = RoutineReference(name=lookup)
                        ref.resolve(self.ff)
                        ref.language = (script, lang)
                        self.ff.addFeature(feature, [ref])

    def convert(self):
        doc = Parser(self._file_or_path).parse()
        if self._font is not None:
            self._glyph_order = self._font.getGlyphOrder()

        self.ff = FontFeatures()
        self._collectStatements(doc)
        return self.ff

    def _glyphName(self, glyph):
        try:
            name = glyph.glyph
        except AttributeError:
            name = glyph
        return name

    def _groupName(self, group):
        try:
            name = group.group
        except AttributeError:
            name = group
        return self.ff.namedClasses[name.lower()]

    def _coverage(self, coverage):
        items = []
        for item in coverage:
            if isinstance(item, VAst.GlyphName):
                items.append([self._glyphName(item)])
            elif isinstance(item, VAst.GroupName):
                items.append(self._groupName(item))
            elif isinstance(item, VAst.Enum):
                items.append(self._enum(item))
            elif isinstance(item, VAst.Range):
                items.append((item.start, item.end))
            else:
                raise NotImplementedError(item)
        return items

    def _enum(self, enum):
        return self._coverage(enum.enum)

    def _context(self, context):
        out = []
        for item in context:
            coverage = self._coverage(item)
            if not isinstance(coverage, (tuple, list)):
                coverage = [coverage]
            out.extend(coverage)
        return out

    def _groupDefinition(self, group):
        name = self._className(group.name)
        glyphs = [self._glyphName(x) for x in group.enum.enum]
        self.ff.namedClasses[group.name.lower()] = glyphs

    def _glyphDefinition(self, glyph):
        try:
            self._glyph_map[glyph.name] = self._glyph_order[glyph.id]
        except TypeError:
            pass

        if glyph.type in ("BASE", "MARK", "LIGATURE", "COMPONENT"):
            self.ff.glyphclasses[glyph.name] = glyph.type.lower()

        if glyph.type == "MARK":
            self._marks.add(glyph.name)
        elif glyph.type == "LIGATURE":
            self._ligatures[glyph.name] = glyph.components

    def _scriptDefinition(self, script):
        stag = script.tag
        for lang in script.langs:
            ltag = lang.tag
            for feature in lang.features:
                lookups = {l.split("\\")[0]: True for l in feature.lookups}
                ftag = feature.tag
                if ftag not in self._features:
                    self._features[ftag] = {}
                if stag not in self._features[ftag]:
                    self._features[ftag][stag] = {}
                assert ltag not in self._features[ftag][stag]
                self._features[ftag][stag][ltag] = lookups.keys()

    def _settingDefinition(self, setting):
        if setting.name.startswith("COMPILER_"):
            self._settings[setting.name] = setting.value
        else:
            log.warning(f"Unsupported setting ignored: {setting.name}")

    def _adjustment(self, adjustment):
        adv, dx, dy, adv_adjust_by, dx_adjust_by, dy_adjust_by = adjustment

        adv_device = adv_adjust_by and adv_adjust_by.items() or None
        dx_device = dx_adjust_by and dx_adjust_by.items() or None
        dy_device = dy_adjust_by and dy_adjust_by.items() or None

        return ValueRecord(
            xPlacement=dx,
            yPlacement=dy,
            xAdvance=adv,
            xPlaDevice=dx_device,
            yPlaDevice=dy_device,
            xAdvDevice=adv_device,
        )

    def _anchor(self, adjustment):
        adv, dx, dy, adv_adjust_by, dx_adjust_by, dy_adjust_by = adjustment
        return (dx or 0, dy or 0)

    def _anchorDefinition(self, anchordef):
        anchorname = anchordef.name
        glyphname = anchordef.glyph_name
        anchor = self._anchor(anchordef.pos)

        if anchorname.startswith("MARK_"):
            name = "_".join(anchorname.split("_")[1:])
            markclass = self._className(name)
            glyph = self._glyphName(glyphname)
            self._markclasses[(glyphname, anchorname)] = markclass
        else:
            if glyphname not in self._anchors:
                self._anchors[glyphname] = {}
            if anchorname not in self._anchors[glyphname]:
                self._anchors[glyphname][anchorname] = {}
            self._anchors[glyphname][anchorname][anchordef.component] = anchor

    def _gposLookup(self, lookup, fealookup):
        statements = fealookup.statements

        pos = lookup.pos
        if isinstance(pos, VAst.PositionAdjustPairDefinition):
            for (idx1, idx2), (pos1, pos2) in pos.adjust_pair.items():
                coverage_1 = pos.coverages_1[idx1 - 1]
                coverage_2 = pos.coverages_2[idx2 - 1]

                # If not both are groups, use “enum pos” otherwise makeotf will
                # fail.
                enumerated = False
                for item in coverage_1 + coverage_2:
                    if not isinstance(item, VAst.GroupName):
                        enumerated = True

                glyphs1 = self._coverage(coverage_1)
                glyphs2 = self._coverage(coverage_2)
                record1 = self._adjustment(pos1)
                record2 = self._adjustment(pos2)
                assert len(glyphs1) == 1
                assert len(glyphs2) == 1
                statements.append(
                    ast.PairPosStatement(
                        glyphs1[0], record1, glyphs2[0], record2, enumerated=enumerated
                    )
                )
        elif isinstance(pos, VAst.PositionAdjustSingleDefinition):
            for a, b in pos.adjust_single:
                glyphs = self._coverage(a)
                record = self._adjustment(b)
                assert len(glyphs) == 1
                statements.append(
                    ast.SinglePosStatement([(glyphs[0], record)], [], [], False)
                )
        elif isinstance(pos, VAst.PositionAttachDefinition):
            anchors = {}
            for marks, classname in pos.coverage_to:
                for mark in marks:
                    # Set actually used mark classes. Basically a hack to get
                    # around the feature file syntax limitation of making mark
                    # classes global and not allowing mark positioning to
                    # specify mark coverage.
                    for name in mark.glyphSet():
                        key = (name, "MARK_" + classname)
                        self._markclasses[key].used = True
                markclass = ast.MarkClass(self._className(classname))
                for base in pos.coverage:
                    for name in base.glyphSet():
                        if name not in anchors:
                            anchors[name] = []
                        if classname not in anchors[name]:
                            anchors[name].append(classname)

            for name in anchors:
                components = 1
                if name in self._ligatures:
                    components = self._ligatures[name]

                marks = []
                for mark in anchors[name]:
                    markclass = ast.MarkClass(self._className(mark))
                    for component in range(1, components + 1):
                        if len(marks) < component:
                            marks.append([])
                        anchor = None
                        if component in self._anchors[name][mark]:
                            anchor = self._anchors[name][mark][component]
                        marks[component - 1].append((anchor, markclass))

                base = self._glyphName(name)
                if name in self._marks:
                    mark = ast.MarkMarkPosStatement(base, marks[0])
                elif name in self._ligatures:
                    mark = ast.MarkLigPosStatement(base, marks)
                else:
                    mark = ast.MarkBasePosStatement(base, marks[0])
                statements.append(mark)
        elif isinstance(pos, VAst.PositionAttachCursiveDefinition):
            # Collect enter and exit glyphs
            enter_coverage = []
            for coverage in pos.coverages_enter:
                for base in coverage:
                    for name in base.glyphSet():
                        enter_coverage.append(name)
            exit_coverage = []
            for coverage in pos.coverages_exit:
                for base in coverage:
                    for name in base.glyphSet():
                        exit_coverage.append(name)

            # Write enter anchors, also check if the glyph has exit anchor and
            # write it, too.
            for name in enter_coverage:
                glyph = self._glyphName(name)
                entry = self._anchors[name]["entry"][1]
                exit = None
                if name in exit_coverage:
                    exit = self._anchors[name]["exit"][1]
                    exit_coverage.pop(exit_coverage.index(name))
                statements.append(ast.CursivePosStatement(glyph, entry, exit))

            # Write any remaining exit anchors.
            for name in exit_coverage:
                glyph = self._glyphName(name)
                exit = self._anchors[name]["exit"][1]
                statements.append(ast.CursivePosStatement(glyph, None, exit))
        else:
            raise NotImplementedError(pos)

    def _gposContextLookup(
        self, lookup, prefix, suffix, ignore, fealookup, targetlookup
    ):
        statements = fealookup.statements

        assert not lookup.reversal

        pos = lookup.pos
        if isinstance(pos, VAst.PositionAdjustPairDefinition):
            for (idx1, idx2), (pos1, pos2) in pos.adjust_pair.items():
                glyphs1 = self._coverage(pos.coverages_1[idx1 - 1])
                glyphs2 = self._coverage(pos.coverages_2[idx2 - 1])
                assert len(glyphs1) == 1
                assert len(glyphs2) == 1
                glyphs = (glyphs1[0], glyphs2[0])

                if ignore:
                    statement = ast.IgnorePosStatement([(prefix, glyphs, suffix)])
                else:
                    lookups = (targetlookup, targetlookup)
                    statement = ast.ChainContextPosStatement(
                        prefix, glyphs, suffix, lookups
                    )
                statements.append(statement)
        elif isinstance(pos, VAst.PositionAdjustSingleDefinition):
            glyphs = [ast.GlyphClass()]
            for a, b in pos.adjust_single:
                glyph = self._coverage(a)
                glyphs[0].extend(glyph)

            if ignore:
                statement = ast.IgnorePosStatement([(prefix, glyphs, suffix)])
            else:
                statement = ast.ChainContextPosStatement(
                    prefix, glyphs, suffix, [targetlookup]
                )
            statements.append(statement)
        elif isinstance(pos, VAst.PositionAttachDefinition):
            glyphs = [ast.GlyphClass()]
            for coverage, _ in pos.coverage_to:
                glyphs[0].extend(self._coverage(coverage))

            if ignore:
                statement = ast.IgnorePosStatement([(prefix, glyphs, suffix)])
            else:
                statement = ast.ChainContextPosStatement(
                    prefix, glyphs, suffix, [targetlookup]
                )
            statements.append(statement)
        else:
            raise NotImplementedError(pos)

    def _gsubLookup(self, lookup, prefix, suffix, ignore, chain, routine):
        sub = lookup.sub
        for key, val in sub.mapping.items():
            if not key or not val:
                path, line, column = sub.location
                log.warning(f"{path}:{line}:{column}: Ignoring empty substitution")
                continue
            statement = None
            glyphs = self._coverage(key)
            replacements = self._coverage(val)
            if ignore:
                chain_context = (prefix, glyphs, suffix)
                statement = ast.IgnoreSubstStatement([chain_context])
            else:
                statement = Substitution(
                    glyphs,
                    replacements,
                    precontext=prefix,
                    postcontext=suffix,
                    lookups=chain
                )
                if isinstance(sub, VAst.SubstitutionReverseChainingSingleDefinition):
                    statement.reversed = True
            routine.rules.append(statement)

    def _lookupDefinition(self, lookup):
        mark_attachement = None
        mark_filtering = None

        flags = 0
        if lookup.direction == "RTL":
            flags |= 1
        if not lookup.process_base:
            flags |= 2
        # FIXME: Does VOLT support this?
        # if not lookup.process_ligatures:
        #     flags |= 4
        if not lookup.process_marks:
            flags |= 8
        elif isinstance(lookup.process_marks, str):
            mark_attachement = self._groupName(lookup.process_marks)
        elif lookup.mark_glyph_set is not None:
            mark_filtering = self._groupName(lookup.mark_glyph_set)
        if "\\" in lookup.name:
            # Merge sub lookups as subtables (lookups named “base\sub”),
            # makeotf/feaLib will issue a warning and ignore the subtable
            # statement if it is not a pairpos lookup, though.
            name = lookup.name.split("\\")[0]
            if name.lower() not in self._lookups:
                fealookup = ast.LookupBlock(self._lookupName(name))
                if lookupflags is not None:
                    fealookup.statements.append(lookupflags)
                fealookup.statements.append(ast.Comment("# " + lookup.name))
            else:
                fealookup = self._lookups[name.lower()]
                fealookup.statements.append(ast.SubtableStatement())
                fealookup.statements.append(ast.Comment("# " + lookup.name))
            self._lookups[name.lower()] = fealookup
        else:
            routine = Routine(name=lookup.name, flags=flags,
                markFilteringSet = mark_filtering,
                markAttachmentSet = mark_attachement)
            self.ff.routines.append(routine)
            self._lookups[lookup.name.lower()] = routine

        contexts = []
        if lookup.context:
            for context in lookup.context:
                prefix = self._context(context.left)
                suffix = self._context(context.right)
                ignore = context.ex_or_in == "EXCEPT_CONTEXT"
                contexts.append([prefix, suffix, ignore, False])
                # It seems that VOLT will create contextual substitution using
                # only the input if there is no other contexts in this lookup.
                if ignore and len(lookup.context) == 1:
                    contexts.append([[], [], False, True])
        else:
            contexts.append([[], [], False, False])

        targetlookup = None
        for prefix, suffix, ignore, chain in contexts:
            if lookup.sub is not None:
                self._gsubLookup(lookup, prefix, suffix, ignore, chain, routine)

            if lookup.pos is not None:
                if self._settings.get("COMPILER_USEEXTENSIONLOOKUPS"):
                    fealookup.use_extension = True
                if prefix or suffix or chain or ignore:
                    if not ignore and targetlookup is None:
                        targetname = self._lookupName(lookup.name + " target")
                        targetlookup = ast.LookupBlock(targetname)
                        fealookup.targets = getattr(fealookup, "targets", [])
                        fealookup.targets.append(targetlookup)
                        self._gposLookup(lookup, targetlookup)
                    self._gposContextLookup(
                        lookup, prefix, suffix, ignore, fealookup, targetlookup
                    )
                # else:
                    # self._gposLookup(lookup, fealookup)


def main(args=None):
    import argparse

    parser = argparse.ArgumentParser(description="Convert VOLT/VTP to feature files.")
    parser.add_argument("input", metavar="INPUT", help="input font/VTP file to process")
    parser.add_argument("featurefile", metavar="FEATUEFILE", help="output feature file")
    parser.add_argument(
        "-q", "--quiet", action="store_true", help="Suppress non-error messages"
    )
    parser.add_argument(
        "--traceback", action="store_true", help="Don’t catch exceptions"
    )

    options = parser.parse_args(args)

    if options.quiet:
        log.setLevel(logging.ERROR)
    logging.basicConfig(format="%(levelname)s: %(message)s")

    file_or_path = options.input
    font = None
    try:
        font = TTFont(file_or_path)
        if "TSIV" in font:
            file_or_path = StringIO(font["TSIV"].data.decode("utf-8"))
        else:
            log.error('"TSIV" table is missing, font was not saved from VOLT?')
            return 1
    except TTLibError:
        pass

    converter = VoltParser(file_or_path, font)
    try:
        ff = converter.convert()
    except NotImplementedError as e:
        if options.traceback:
            raise
        location = getattr(e.args[0], "location", None)
        message = f'"{e}" is not supported'
        if location:
            path, line, column = location
            log.error(f"{path}:{line}:{column}: {message}")
        else:
            log.error(message)
        return 1
    print(ff.asFea())


if __name__ == "__main__":
    import sys

    sys.exit(main())



from fontFeatures import Substitution
from lxml import etree

import unittest


class TestSubstitution(unittest.TestCase):
    def roundTrip(self, thing):
        rt = thing.__class__.fromXML(thing.toXML())
        self.assertEqual(rt.asFea(), thing.asFea())

    def test_single(self):
        s = Substitution(["a"], ["b"])
        self.assertEqual(s.asFea(), "sub a by b;")
        self.assertEqual(s.involved_glyphs, set(["a", "b"]))
        self.assertEqual(etree.tostring(s.toXML()), "<substitution><from><slot><glyph>a</glyph></slot></from><to><slot><glyph>b</glyph></slot></to></substitution>".encode("utf-8"))
        self.roundTrip(s)

    def test_single_classes(self):
        s = Substitution([["a", "b"]], [["c", "d"]])
        self.assertEqual(s.asFea(), "sub [a b] by [c d];")
        self.assertEqual(s.involved_glyphs, set(["a", "b", "c", "d"]))
        self.roundTrip(s)

    def test_ligature(self):
        s = Substitution(["a", "b"], ["c"])
        self.assertEqual(s.asFea(), "sub a b by c;")
        self.assertEqual(s.involved_glyphs, set(["a", "b", "c"]))
        self.roundTrip(s)

    def test_ligature_expansion(self):
        s = Substitution([["f", "f.ss01"], ["i", "i.ss01"]], [["f_i", "f_i.ss01"]])
        self.assertEqual(s.asFea(), "    sub f i by f_i;\n    sub f.ss01 i.ss01 by f_i.ss01;\n")
        self.assertEqual(s.involved_glyphs, set(["f", "i", "f_i", "f.ss01", "i.ss01", "f_i.ss01"]))
        self.roundTrip(s)

    def test_multiple(self):
        s = Substitution(["a"], ["b", "c"])
        self.assertEqual(s.asFea(), "sub a by b c;")
        self.assertEqual(s.involved_glyphs, set(["a", "b", "c"]))
        self.roundTrip(s)

    def test_multiple_expansion(self):
        s = Substitution([["a", "b"]], [["a", "b"], "c"])
        self.assertEqual(s.asFea(), "    sub a by a c;\n    sub b by b c;\n")
        self.assertEqual(s.involved_glyphs, set(["a", "b", "c"]))
        self.roundTrip(s)

    def test_multiple_expansion_middle(self):
        s = Substitution([["aa", "bb"]], ["d", ["aa", "bb"], "c"])
        self.assertEqual(s.asFea(), "    sub aa by d aa c;\n    sub bb by d bb c;\n")
        self.assertEqual(s.involved_glyphs, set(["aa", "bb", "c", "d"]))
        self.roundTrip(s)

    def test_alternate(self):
        s = Substitution(["a"], [["b", "c"]])
        self.assertEqual(s.asFea(), "sub a from [b c];")
        self.assertEqual(s.involved_glyphs, set(["a", "b", "c"]))
        self.roundTrip(s)
